@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import contextlib
 import datetime
 import hashlib
 import json
@@ -12,14 +13,16 @@ from collections import defaultdict
 from itertools import chain
 from typing import Collection, Iterable, Mapping, Sequence, cast
 
-import click
 import tomlkit
 from resolvelib.reporters import BaseReporter
 from resolvelib.resolvers import ResolutionImpossible, ResolutionTooDeep, Resolver
+from tomlkit.items import Array
 
-from pdm import signals, termui
+from pdm import termui
+from pdm.cli.hooks import HookManager
 from pdm.cli.utils import (
     check_project_file,
+    fetch_hashes,
     find_importable_files,
     format_lockfile,
     format_resolution_impossible,
@@ -56,8 +59,10 @@ def do_lock(
     requirements: list[Requirement] | None = None,
     dry_run: bool = False,
     refresh: bool = False,
+    hooks: HookManager | None = None,
 ) -> dict[str, Candidate]:
     """Performs the locking process and update lockfile."""
+    hooks = hooks or HookManager(project)
     check_project_file(project)
     if refresh:
         locked_repo = project.locked_repository
@@ -67,12 +72,12 @@ def do_lock(
         with project.core.ui.open_spinner("Re-calculating hashes..."):
             for key, candidate in locked_repo.packages.items():
                 reqs, python_requires, summary = locked_repo.candidate_info[key]
-                candidate.hashes = repo.get_hashes(candidate)
                 candidate.summary = summary
                 candidate.requires_python = python_requires
                 ident = cast(str, key[0])
                 mapping[ident] = candidate
                 dependencies[ident] = list(map(parse_requirement, reqs))
+            fetch_hashes(repo, mapping)
             lockfile = format_lockfile(project, mapping, dependencies)
         project.write_lockfile(lockfile)
         return mapping
@@ -87,37 +92,37 @@ def do_lock(
     with ui.logging("lock"):
         # The context managers are nested to ensure the spinner is stopped before
         # any message is thrown to the output.
-        with ui.open_spinner(title="Resolving dependencies", spinner="dots") as spin:
-            reporter = project.get_reporter(requirements, tracked_names, spin)
-            resolver: Resolver = project.core.resolver_class(provider, reporter)
-            signals.pre_lock.send(project, requirements=requirements, dry_run=dry_run)
-            try:
+        try:
+            with ui.open_spinner(title="Resolving dependencies") as spin:
+                reporter = project.get_reporter(requirements, tracked_names, spin)
+                resolver: Resolver = project.core.resolver_class(provider, reporter)
+                hooks.try_emit("pre_lock", requirements=requirements, dry_run=dry_run)
                 mapping, dependencies = resolve(
                     resolver,
                     requirements,
                     project.environment.python_requires,
                     resolve_max_rounds,
                 )
-            except ResolutionTooDeep:
-                spin.fail(f"{termui.Emoji.LOCK} Lock failed")
-                ui.echo(
-                    "The dependency resolution exceeds the maximum loop depth of "
-                    f"{resolve_max_rounds}, there may be some circular dependencies "
-                    "in your project. Try to solve them or increase the "
-                    f"{termui.green('`strategy.resolve_max_rounds`')} config.",
-                    err=True,
-                )
-                raise
-            except ResolutionImpossible as err:
-                spin.fail(f"{termui.Emoji.LOCK} Lock failed")
-                ui.echo(format_resolution_impossible(err), err=True)
-                raise ResolutionImpossible("Unable to find a resolution") from None
-            else:
-                data = format_lockfile(project, mapping, dependencies)
-                spin.succeed(f"{termui.Emoji.LOCK} Lock successful")
-            signals.post_lock.send(project, resolution=mapping, dry_run=dry_run)
-
-    project.write_lockfile(data, write=not dry_run)
+                fetch_hashes(provider.repository, mapping)
+        except ResolutionTooDeep:
+            ui.echo(f"{termui.Emoji.LOCK} Lock failed", err=True)
+            ui.echo(
+                "The dependency resolution exceeds the maximum loop depth of "
+                f"{resolve_max_rounds}, there may be some circular dependencies "
+                "in your project. Try to solve them or increase the "
+                f"[green]`strategy.resolve_max_rounds`[/] config.",
+                err=True,
+            )
+            raise
+        except ResolutionImpossible as err:
+            ui.echo(f"{termui.Emoji.LOCK} Lock failed", err=True)
+            ui.echo(format_resolution_impossible(err), err=True)
+            raise ResolutionImpossible("Unable to find a resolution") from None
+        else:
+            data = format_lockfile(project, mapping, dependencies)
+            ui.echo(f"{termui.Emoji.LOCK} Lock successful")
+            project.write_lockfile(data, write=not dry_run)
+            hooks.try_emit("post_lock", resolution=mapping, dry_run=dry_run)
 
     return mapping
 
@@ -143,6 +148,7 @@ def resolve_candidates_from_lockfile(
                 project.environment.python_requires,
                 resolve_max_rounds,
             )
+            fetch_hashes(provider.repository, mapping)
     return mapping
 
 
@@ -151,19 +157,19 @@ def check_lockfile(project: Project, raise_not_exist: bool = True) -> str | None
     if not project.lockfile_file.exists():
         if raise_not_exist:
             raise ProjectError("Lock file does not exist, nothing to install")
-        project.core.ui.echo("Lock file does not exist", fg="yellow", err=True)
+        project.core.ui.echo("Lock file does not exist", style="yellow", err=True)
         return "all"
     elif not project.is_lockfile_compatible():
         project.core.ui.echo(
             "Lock file version is not compatible with PDM, installation may fail",
-            fg="yellow",
+            style="yellow",
             err=True,
         )
         return "all"
     elif not project.is_lockfile_hash_match():
         project.core.ui.echo(
             "Lock file hash doesn't match pyproject.toml, packages may be outdated",
-            fg="yellow",
+            style="yellow",
             err=True,
         )
         return "reuse"
@@ -183,8 +189,11 @@ def do_sync(
     no_editable: bool | Collection[str] = False,
     no_self: bool = False,
     reinstall: bool = False,
+    only_keep: bool = False,
+    hooks: HookManager | None = None,
 ) -> None:
     """Synchronize project"""
+    hooks = hooks or HookManager(project)
     if requirements is None:
         groups = translate_groups(project, default, dev, groups or ())
         requirements = []
@@ -204,10 +213,11 @@ def do_sync(
         install_self=not no_self and "default" in groups and bool(project.name),
         use_install_cache=project.config["install.cache"],
         reinstall=reinstall,
+        only_keep=only_keep,
     )
-    signals.pre_install.send(project, candidates=candidates, dry_run=dry_run)
+    hooks.try_emit("pre_install", candidates=candidates, dry_run=dry_run)
     handler.synchronize()
-    signals.post_install.send(project, candidates=candidates, dry_run=dry_run)
+    hooks.try_emit("post_install", candidates=candidates, dry_run=dry_run)
 
 
 def do_add(
@@ -224,8 +234,10 @@ def do_add(
     no_self: bool = False,
     dry_run: bool = False,
     prerelease: bool = False,
+    hooks: HookManager | None = None,
 ) -> None:
     """Add packages and install"""
+    hooks = hooks or HookManager(project)
     check_project_file(project)
     if not editables and not packages:
         raise PdmUsageError("Must specify at least one package or editable package.")
@@ -235,6 +247,15 @@ def do_add(
         group = "dev" if dev else "default"
     tracked_names: set[str] = set()
     requirements: dict[str, Requirement] = {}
+    if (
+        group == "default"
+        or not dev
+        and group not in project.tool_settings.get("dev-dependencies", {})
+    ):
+        if editables:
+            raise PdmUsageError(
+                "Cannot add editables to the default or optional dependency group"
+            )
     for r in [parse_requirement(line, True) for line in editables] + [
         parse_requirement(line) for line in packages
     ]:
@@ -244,7 +265,7 @@ def do_add(
         requirements[key] = r
     project.core.ui.echo(
         f"Adding packages to {group} {'dev-' if dev else ''}dependencies: "
-        + ", ".join(termui.green(r.as_line(), bold=True) for r in requirements.values())
+        + ", ".join(f"[bold green]{r.as_line()}[/]" for r in requirements.values())
     )
     all_dependencies = project.all_dependencies
     group_deps = all_dependencies.setdefault(group, {})
@@ -253,15 +274,15 @@ def do_add(
             req.specifier = get_specifier("")
     group_deps.update(requirements)
     reqs = [r for deps in all_dependencies.values() for r in deps.values()]
-    resolved = do_lock(project, strategy, tracked_names, reqs, dry_run=dry_run)
+    resolved = do_lock(
+        project, strategy, tracked_names, reqs, dry_run=dry_run, hooks=hooks
+    )
 
     # Update dependency specifiers and lockfile hash.
     deps_to_update = group_deps if unconstrained else requirements
     save_version_specifiers({group: deps_to_update}, resolved, save)
     if not dry_run:
-        project.add_dependencies(
-            deps_to_update, group, dev, replace_editable=no_editable
-        )
+        project.add_dependencies(deps_to_update, group, dev)
         project.write_lockfile(project.lockfile, False)
 
     if sync:
@@ -273,6 +294,7 @@ def do_add(
             no_self=no_self,
             requirements=list(group_deps.values()),
             dry_run=dry_run,
+            hooks=hooks,
         )
 
 
@@ -292,8 +314,10 @@ def do_update(
     no_editable: bool = False,
     no_self: bool = False,
     prerelease: bool = False,
+    hooks: HookManager | None = None,
 ) -> None:
     """Update specified packages or all packages"""
+    hooks = hooks or HookManager(project)
     check_project_file(project)
     if len(packages) > 0 and (top or len(groups) > 1 or not default):
         raise PdmUsageError(
@@ -314,25 +338,24 @@ def do_update(
         dependencies = all_dependencies[group]
         for name in packages:
             matched_name = next(
-                filter(
-                    lambda k: normalize_name(strip_extras(k)[0])
-                    == normalize_name(name),
-                    dependencies.keys(),
+                (
+                    k
+                    for k in dependencies
+                    if normalize_name(strip_extras(k)[0]) == normalize_name(name)
                 ),
                 None,
             )
             if not matched_name:
                 raise ProjectError(
-                    "{} does not exist in {} {}dependencies.".format(
-                        termui.green(name, bold=True), group, "dev-" if dev else ""
-                    )
+                    f"[bold green]{name}[/] does not exist in {group} "
+                    f"{'dev-' if dev else ''}dependencies."
                 )
             dependencies[matched_name].prerelease = prerelease
             updated_deps[group][matched_name] = dependencies[matched_name]
         project.core.ui.echo(
             "Updating packages: {}.".format(
                 ", ".join(
-                    termui.green(v, bold=True)
+                    f"[bold green]{v}[/]"
                     for v in chain.from_iterable(updated_deps.values())
                 )
             )
@@ -348,6 +371,7 @@ def do_update(
         chain.from_iterable(updated_deps.values()),
         reqs,
         dry_run=dry_run,
+        hooks=hooks,
     )
     if sync or dry_run:
         do_sync(
@@ -363,6 +387,7 @@ def do_update(
             else None,
             no_editable=no_editable,
             no_self=no_self,
+            hooks=hooks,
         )
     if unconstrained and not dry_run:
         # Need to update version constraints
@@ -382,8 +407,10 @@ def do_remove(
     no_editable: bool = False,
     no_self: bool = False,
     dry_run: bool = False,
+    hooks: HookManager | None = None,
 ) -> None:
     """Remove packages from working set and pyproject.toml"""
+    hooks = hooks or HookManager(project)
     check_project_file(project)
     if not packages:
         raise PdmUsageError("Must specify at least one package to remove.")
@@ -395,25 +422,24 @@ def do_remove(
     deps = project.get_pyproject_dependencies(group, dev)
     project.core.ui.echo(
         f"Removing packages from {group} {'dev-' if dev else ''}dependencies: "
-        + ", ".join(str(termui.green(name, bold=True)) for name in packages)
+        + ", ".join(f"[bold green]{name}[/]" for name in packages)
     )
     for name in packages:
         req = parse_requirement(name)
         matched_indexes = sorted(
-            (i for i, r in enumerate(deps) if req.matches(r, False)), reverse=True
+            (i for i, r in enumerate(deps) if req.matches(r)), reverse=True
         )
         if not matched_indexes:
             raise ProjectError(
-                "{} does not exist in {} dependencies.".format(
-                    termui.green(name, bold=True), group
-                )
+                f"[bold green]{name}[/] does not exist in {group} dependencies."
             )
         for i in matched_indexes:
             del deps[i]
+    cast(Array, deps).multiline(True)
 
     if not dry_run:
         project.write_pyproject()
-    do_lock(project, "reuse", dry_run=dry_run)
+    do_lock(project, "reuse", dry_run=dry_run, hooks=hooks)
     if sync:
         do_sync(
             project,
@@ -423,6 +449,7 @@ def do_remove(
             no_editable=no_editable,
             no_self=no_self,
             dry_run=dry_run,
+            hooks=hooks,
         )
 
 
@@ -434,7 +461,7 @@ def do_list(
     json: bool = False,
 ) -> None:
     """Display a list of packages installed in the local packages directory."""
-    from pdm.cli.utils import build_dependency_graph, format_dependency_graph
+    from pdm.cli.utils import build_dependency_graph, show_dependency_graph
 
     check_project_file(project)
     working_set = project.environment.get_working_set()
@@ -442,9 +469,7 @@ def do_list(
         dep_graph = build_dependency_graph(
             working_set, project.environment.marker_environment
         )
-        project.core.ui.echo(
-            format_dependency_graph(project, dep_graph, reverse=reverse, json=json)
-        )
+        show_dependency_graph(project, dep_graph, reverse=reverse, json=json)
     else:
         if reverse:
             raise PdmUsageError("--reverse must be used with --graph")
@@ -468,7 +493,7 @@ def do_list(
             project.core.ui.echo("\n".join(reqs))
             return
         rows = [
-            (termui.green(k, bold=True), termui.yellow(v.version), get_dist_location(v))
+            (f"[b green]{k}[/]", f"[yellow]{v.version}[/]", get_dist_location(v))
             for k, v in sorted(working_set.items())
         ]
         project.core.ui.display_columns(rows, ["Package", "Version", "Location"])
@@ -481,10 +506,13 @@ def do_build(
     dest: str = "dist",
     clean: bool = True,
     config_settings: Mapping[str, str] | None = None,
+    hooks: HookManager | None = None,
 ) -> None:
     """Build artifacts for distribution."""
     from pdm.builders import SdistBuilder, WheelBuilder
 
+    hooks = hooks or HookManager(project)
+
     if project.is_global:
         raise ProjectError("Not allowed to build based on the global project.")
     if not wheel and not sdist:
@@ -494,7 +522,7 @@ def do_build(
         dest = project.root.joinpath(dest).as_posix()
     if clean:
         shutil.rmtree(dest, ignore_errors=True)
-    signals.pre_build.send(project, dest=dest, config_settings=config_settings)
+    hooks.try_emit("pre_build", dest=dest, config_settings=config_settings)
     artifacts: list[str] = []
     with project.core.ui.logging("build"):
         if sdist:
@@ -511,9 +539,7 @@ def do_build(
             )
             project.core.ui.echo(f"Built wheel at {loc}")
             artifacts.append(loc)
-    signals.post_build.send(
-        project, artifacts=artifacts, config_settings=config_settings
-    )
+    hooks.try_emit("post_build", artifacts=artifacts, config_settings=config_settings)
 
 
 def do_init(
@@ -525,8 +551,10 @@ def do_init(
     author: str = "",
     email: str = "",
     python_requires: str = "",
+    hooks: HookManager | None = None,
 ) -> None:
     """Bootstrap the project and create a pyproject.toml"""
+    hooks = hooks or HookManager(project)
     data = {
         "project": {
             "name": name,
@@ -537,7 +565,7 @@ def do_init(
             "dependencies": make_array([], True),
         },
         "build-system": {
-            "requires": ["pdm-pep517>=0.12.0"],
+            "requires": ["pdm-pep517>=1.0.0"],
             "build-backend": "pdm.pep517.api",
         },
     }
@@ -556,7 +584,7 @@ def do_init(
         project._pyproject["project"] = data["project"]  # type: ignore
         project._pyproject["build-system"] = data["build-system"]  # type: ignore
     project.write_pyproject()
-    signals.post_init.send(project)
+    hooks.try_emit("post_init")
 
 
 def do_use(
@@ -564,15 +592,21 @@ def do_use(
     python: str = "",
     first: bool = False,
     ignore_remembered: bool = False,
-) -> None:
+    ignore_requires_python: bool = False,
+    hooks: HookManager | None = None,
+) -> PythonInfo:
     """Use the specified python version and save in project config.
     The python can be a version string or interpreter path.
     """
+    hooks = hooks or HookManager(project)
+
     if python:
         python = python.strip()
 
     def version_matcher(py_version: PythonInfo) -> bool:
-        return project.python_requires.contains(str(py_version.version), True)
+        return ignore_requires_python or project.python_requires.contains(
+            str(py_version.version), True
+        )
 
     if not project.cache_dir.exists():
         project.cache_dir.mkdir(parents=True)
@@ -581,52 +615,55 @@ def do_use(
     )
     selected_python: PythonInfo | None = None
     if python and not ignore_remembered:
-        if use_cache.has_key(python):
+        if python in use_cache:
             path = use_cache.get(python)
             cached_python = PythonInfo.from_path(path)
             if not cached_python.valid:
                 project.core.ui.echo(
                     f"The last selection is corrupted. {path!r}",
-                    fg="red",
+                    style="red",
                     err=True,
                 )
             elif version_matcher(cached_python):
                 project.core.ui.echo(
                     "Using the last selection, add '-i' to ignore it.",
-                    fg="yellow",
+                    style="yellow",
                     err=True,
                 )
                 selected_python = cached_python
 
     if selected_python is None:
         found_interpreters = list(dict.fromkeys(project.find_interpreters(python)))
-        matching_interperters = list(filter(version_matcher, found_interpreters))
+        matching_interpreters = list(filter(version_matcher, found_interpreters))
         if not found_interpreters:
-            raise NoPythonVersion("Python interpreter is not found on the system.")
-        if not matching_interperters:
+            raise NoPythonVersion(
+                f"No Python interpreter matching [green]{python}[/] is found."
+            )
+        if not matching_interpreters:
             project.core.ui.echo("Interpreters found but not matching:", err=True)
             for py in found_interpreters:
                 project.core.ui.echo(f"  - {py.executable} ({py.identifier})", err=True)
             raise NoPythonVersion(
                 "No python is found meeting the requirement "
-                f"{termui.green('python' + str(project.python_requires))}"
+                f"[green]python {str(project.python_requires)}[/]"
             )
-        if first or len(matching_interperters) == 1:
-            selected_python = matching_interperters[0]
+        if first or len(matching_interpreters) == 1:
+            selected_python = matching_interpreters[0]
         else:
             project.core.ui.echo("Please enter the Python interpreter to use")
-            for i, py_version in enumerate(matching_interperters):
+            for i, py_version in enumerate(matching_interpreters):
                 project.core.ui.echo(
-                    f"{i}. {termui.green(str(py_version.executable))} "
+                    f"{i}. [green]{str(py_version.executable)}[/] "
                     f"({py_version.identifier})"
                 )
-            selection = click.prompt(
-                "Please select:",
-                type=click.Choice([str(i) for i in range(len(matching_interperters))]),
+            selection = termui.ask(
+                "Please select",
                 default="0",
+                prompt_type=int,
+                choices=[str(i) for i in range(len(matching_interpreters))],
                 show_choices=False,
             )
-            selected_python = matching_interperters[int(selection)]
+            selected_python = matching_interpreters[int(selection)]
         if python:
             use_cache.set(python, selected_python.path.as_posix())
 
@@ -636,10 +673,9 @@ def do_use(
 
     old_python = project.python if "python.path" in project.config else None
     project.core.ui.echo(
-        "Using Python interpreter: {} ({})".format(
-            termui.green(str(selected_python.executable)),
-            selected_python.identifier,
-        )
+        "Using Python interpreter: "
+        f"[green]{str(selected_python.executable)}[/] "
+        f"({selected_python.identifier})"
     )
     project.python = selected_python
     if (
@@ -647,8 +683,10 @@ def do_use(
         and old_python.path != selected_python.path
         and not project.environment.is_global
     ):
-        project.core.ui.echo(termui.cyan("Updating executable scripts..."))
+        project.core.ui.echo("Updating executable scripts...", style="cyan")
         project.environment.update_shebangs(selected_python.executable.as_posix())
+    hooks.try_emit("post_use", python=selected_python)
+    return selected_python
 
 
 def do_import(
@@ -682,7 +720,13 @@ def do_import(
 
     if "tool" not in pyproject or "pdm" not in pyproject["tool"]:  # type: ignore
         pyproject.setdefault("tool", {})["pdm"] = tomlkit.table()
-
+    if "build" in pyproject["tool"]["pdm"] and isinstance(
+        pyproject["tool"]["pdm"]["build"], str
+    ):
+        pyproject["tool"]["pdm"]["build"] = {
+            "setup-script": pyproject["tool"]["pdm"]["build"],
+            "run-setuptools": True,
+        }
     if "project" not in pyproject:
         pyproject.add("project", tomlkit.table())  # type: ignore
         pyproject["project"].add(  # type: ignore
@@ -695,7 +739,7 @@ def do_import(
     merge_dictionary(pyproject["project"], project_data)  # type: ignore
     merge_dictionary(pyproject["tool"]["pdm"], settings)  # type: ignore
     pyproject["build-system"] = {
-        "requires": ["pdm-pep517"],
+        "requires": ["pdm-pep517>=1.0.0"],
         "build-backend": "pdm.pep517.api",
     }
     project.pyproject = cast(dict, pyproject)
@@ -708,20 +752,18 @@ def ask_for_import(project: Project) -> None:
     if not importable_files:
         return
     project.core.ui.echo(
-        termui.cyan("Found following files from other formats that you may import:")
+        "Found following files from other formats that you may import:", style="cyan"
     )
     for i, (key, filepath) in enumerate(importable_files):
-        project.core.ui.echo(f"{i}. {termui.green(filepath.as_posix())} ({key})")
+        project.core.ui.echo(f"{i}. [green]{filepath.as_posix()}[/] ({key})")
     project.core.ui.echo(
-        "{}. {}".format(
-            len(importable_files),
-            termui.yellow("don't do anything, I will import later."),
-        )
+        f"{len(importable_files)}. [yellow]don't do anything, I will import later.[/]"
     )
-    choice = click.prompt(
-        "Please select:",
-        type=click.Choice([str(i) for i in range(len(importable_files) + 1)]),
-        show_default=False,
+    choice = termui.ask(
+        "Please select",
+        prompt_type=int,
+        choices=[str(i) for i in range(len(importable_files) + 1)],
+        show_choices=False,
     )
     if int(choice) == len(importable_files):
         return
@@ -738,16 +780,14 @@ def print_pep582_command(ui: termui.UI, shell: str = "AUTO") -> None:
             set_env_in_reg("PYTHONPATH", PEP582_PATH)
         except PermissionError:
             ui.echo(
-                termui.red(
-                    "Permission denied, please run the terminal as administrator."
-                ),
+                "Permission denied, please run the terminal as administrator.",
+                style="red",
                 err=True,
             )
         ui.echo(
-            termui.green(
-                "The environment variable has been saved, "
-                "please restart the session to take effect."
-            )
+            "The environment variable has been saved, "
+            "please restart the session to take effect.",
+            style="green",
         )
         return
     lib_path = PEP582_PATH.replace("'", "\\'")
@@ -798,26 +838,27 @@ def print_pep582_command(ui: termui.UI, shell: str = "AUTO") -> None:
 
 def get_latest_version(project: Project) -> str | None:
     """Get the latest version of PDM from PyPI, cache for 7 days"""
-    from pdm.utils import get_finder
-
     cache_key = hashlib.sha224(sys.executable.encode()).hexdigest()
     cache_file = project.cache("self-check") / cache_key
-    if cache_file.exists():
+    state = {}
+    with contextlib.suppress(OSError):
         state = json.loads(cache_file.read_text())
-    else:
-        state = {}
     current_time = datetime.datetime.utcnow().timestamp()
     if (
         state.get("last-check")
         and current_time - state["last-check"] < 60 * 60 * 24 * 7
     ):
         return cast(str, state["latest-version"])
-    candidate = get_finder([], project.cache_dir.as_posix()).find_best_candidate("pdm")
-    if not candidate.best_candidate:
+    with project.environment.get_finder(
+        [project.default_source], ignore_compatibility=True
+    ) as finder:
+        candidate = finder.find_best_match("pdm").best
+    if not candidate:
         return None
-    latest_version = str(candidate.best_candidate.version)
+    latest_version = str(candidate.version)
     state.update({"latest-version": latest_version, "last-check": current_time})
-    cache_file.write_text(json.dumps(state))
+    with contextlib.suppress(OSError):
+        cache_file.write_text(json.dumps(state))
     return latest_version
 
 
@@ -841,23 +882,23 @@ def check_update(project: Project) -> None:
     ):
         return
     if is_pipx_installation():  # pragma: no cover
-        install_command = "$ pipx upgrade pdm"
+        install_command = "pipx upgrade pdm"
     elif is_scoop_installation():  # pragma: no cover
-        install_command = "$ scoop update pdm"
+        install_command = "scoop update pdm"
     elif is_homebrew_installation():  # pragma: no cover
-        install_command = "$ brew upgrade pdm"
+        install_command = "brew upgrade pdm"
     else:
-        install_command = f"$ {quote(sys.executable)} -m pip install -U pdm"
+        install_command = f"{quote(sys.executable)} -m pip install -U pdm"
 
-    disable_command = "$ pdm config check_update false"
+    disable_command = "pdm config check_update false"
 
     message = [
-        termui.blue(f"\nPDM {termui.cyan(this_version)}"),
-        termui.blue(f" is installed, while {termui.cyan(latest_version)}"),
-        termui.blue(" is available.\n"),
-        termui.blue(f"Please run {termui.green(install_command, bold=True)}"),
-        termui.blue(" to upgrade.\n"),
-        termui.blue(f"Run {termui.green(disable_command, bold=True)}"),
-        termui.blue(" to disable the check."),
+        f"\nPDM [cyan]{this_version}[/]",
+        f" is installed, while [cyan]{latest_version}[/]",
+        " is available.\n",
+        f"Please run [bold green]`{install_command}`[/]",
+        " to upgrade.\n",
+        f"Run [bold green]`{disable_command}`[/]",
+        " to disable the check.",
     ]
-    project.core.ui.echo("".join(message), err=True)
+    project.core.ui.echo("".join(message), err=True, style="blue")
